// idlcomp/languages/c++/interface-caller.cc -- caller-side interface definitions
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <targets.h>
#include "c++.h"

struct callback_data {
	CPPFile *file;
	Interface *iface;
	bool prototype;
	int count;
};

void CPPFile::output_iface_tbl_entry(Interface *iface, Interface *super)
{
	file << indent << "::System::RunTime::IFaceTable "
	     << *super->get_fq_name()->flatten("_IDLNS_") << ";\n";
}

void CPPFile::tbl_callback(Interface *super, void *arg)
{
	const callback_data *data = static_cast<const callback_data *>(arg);
	data->file->output_iface_tbl_entry(data->iface, super);
}

void CPPFile::output_iface_tbl(Interface *iface)
{
	file << indent << "struct iface_table {\n";

	indent.indent_level++;

	output_iface_tbl_entry(iface, iface);

	callback_data data = { this, iface };
	
	iface->for_each_super<tbl_callback>(&data);
	file << indent << "unsigned char *_end;\n";

	indent.indent_level--;
	
	file << indent << "} iface_table;\n\n";
}

void cpp_output_one_param(ostream &file, Param *p,
                          bool is_server, bool is_copy)
{
	bool is_mutable = p->def.flags.field.Shared ||
	                  p->def.flags.field.Push ||
	                  (is_server && 
	                   (p->def.flags.field.Inline ||
	                    is_copy));
 
	if (dynamic_cast<Interface *>(*p->type)) {
		cpp_output_name(file, p->type);
		file << ' ';
	} else if (p->type) {
		cpp_output_type(file, p->type, p->is_array(), is_mutable);
	} else {
		cpp_output_type(file, p->def.basictype, is_mutable);
	}

	if (p->def.flags.field.Out)
		file << '*';
	
	if (dynamic_cast<Struct *>(*p->type) && !p->is_inline())
		file << '*';

	file << **p->name;
}

void CPPFile::output_one_method(Interface *iface, Method *m, 
                                bool prototype, bool retval)
{
	const String &nsname = get_definition_name(iface);

	file << '\n' << indent << "inline ";
	
	if (!retval)
		file << "void ";
	
	if (!prototype)
		file << get_definition_name(iface) << "::";
	
	file << **m->name << '(';

	if (prototype)
		indent.align_spaces = m->name->length() + 8;
	else
		indent.align_spaces = m->name->length() + nsname.length() + 10;

	if (!retval)
		indent.align_spaces += 5;
	
	bool first = true;
	for (Method::entries_iterator i = m->entries_begin();
	     i != m->entries_end(); ++i)
	{
		if (!first)
			file << ",\n" << indent;
		
		first = false;
		
		cpp_output_one_param(file, *i, false);
	}
	
	file << ")";
	indent.align_spaces = 0;

	if (prototype) {
		file << ";\n";
		return;
	}
	
	file << '\n' << indent << "{\n";
	downscope();

	// FIXME: It'd be nice to skip the NULL test in this cast,
	// as 

	file << indent << "_ptr->info->methods." << **m->name
	     << "(static_cast< ";

	cpp_output_name(file, m->get_ns());

	file << ">(*this)._ptr";

	indent.align_spaces = m->name->length() + 21;
	
	for (Method::entries_iterator i = m->entries_begin();
	     i != m->entries_end(); ++i)
	{
		Param *p = *i;
		file << ",\n" << indent << **p->name;
	}
	
	file << ");\n";
	indent.align_spaces = 0;

	upscope();
	file << indent << "}\n";
}

void CPPFile::output_methods(Interface *iface, Interface *super, bool prototype)
{
	for (Interface::methods_iterator m = super->methods_begin();
	     m != super->methods_end(); ++m)
	{
		output_one_method(iface, *m, prototype, false);
		
		/*
		if (method has at least one out parameter)
			output_one_method(iface, *m, prototype, true);
		*/
	}
}

void CPPFile::output_one_method_ptr(Method *m, Interface *iface)
{
	Interface *super = static_cast<Interface *>(m->get_ns());
	extra_newline();

	file << indent << "void (*" << **m->name << ")(void *_this";
	indent.align_spaces = m->name->length() + 9;
	
	for (Method::entries_iterator p = m->entries_begin();
	     p != m->entries_end(); ++p)
	{
		file << ",\n" << indent;
		cpp_output_one_param(file, *p, false);
	}
	
	file << ");\n";
	indent.align_spaces = 0;
}

void CPPFile::output_method_ptrs_callback(Interface *super, void *arg)
{
	const callback_data *data = static_cast<const callback_data *>(arg);

	for (Interface::methods_iterator m = super->methods_begin();
	     m != super->methods_end(); ++m)
		data->file->output_one_method_ptr(*m, data->iface);
}

void CPPFile::output_method_ptrs(Interface *iface)
{
	file << indent << "struct methods {\n";
	downscope();
	
	callback_data data = { this, iface };

	output_method_ptrs_callback(iface, &data);
	iface->for_each_super<output_method_ptrs_callback>(&data);

	upscope();
	file << indent << "} methods;\n";
}

void CPPFile::output_internal(Interface *iface)
{
	file << indent << "struct info_type {\n";

	indent.indent_level++;
	int super_count = 0;
	
	if (iface->supers_empty()) {
		// No supers.  If it's System.Object, output the top-level
		// info_type.  Otherwise, inherit from System.Object.
		
		if (iface != System_Object) {
			yyerrorf("Interface \"%s\" has no superinterface and "
			         "is not \"System.Object\".", 
			         iface->get_fq_name()->flatten()->c_str());

			throw UserError();
		}
			
		file << indent << "ptrdiff_t concrete;\n"
		     << indent << "::System::RunTime::IFaceTable "
		                  "*concrete_IFaceTable;\n";
	} else {
		Interface *super = *iface->supers_begin();

		file << indent;
		cpp_output_name(file, super, "_i_");
		file << "::info_type parent;\n";
	}

	file << '\n';

	output_iface_tbl(iface);
	output_method_ptrs(iface);
	
	indent.indent_level--;
	
	file << indent << "};\n\n"
	     << indent << "const info_type *const info;\n";

	file << '\n' << indent << "_i_" << **iface->name << '(';
	indent.align_spaces = iface->name->length() + 4;
	cpp_output_name(file, iface, "_i_");

	file << "::info_type *INFO) :\n"
	     << indent << "info(INFO)\n";
	
	indent.align_spaces = 0;
	file << indent << "{}\n";
}

void CPPFile::output_downcast_proto(Interface *iface, Interface *super)
{
	file << '\n' << indent << "static inline " << **iface->name << " downcast(";
	cpp_output_type(file, super, false, false);
	file << "oldptr);\n";
}

void CPPFile::output_downcast(Interface *iface, Interface *super)
{
	const String &name = get_definition_name(iface);
	const String &iname = get_definition_name(iface, "_i_");

	file << '\n' << indent << "inline " << name << ' '
	     << name << "::downcast(";
	cpp_output_type(file, super, false, false);
	file << "oldptr)";
	
	file << "\n" << indent << "{\n";

	indent.indent_level++;
	
	file << indent << "::System::_i_Object *_llptr = \n"
	     << indent << "\treinterpret_cast< ::System::_i_Object *>(oldptr._ptr);\n"
	     << indent << "return " << name << "(reinterpret_cast< "
	     << iname << " *>\n"
	     << indent << "\t(::System::RunTime::downcast(_llptr, &"
	     << name << "_ns::_guid)));\n";
	
	indent.indent_level--;
	
	file << indent << "}\n";
}

void CPPFile::output_upcast_proto(Interface *iface, Interface *super)
{
	const String &supername = *super->get_fq_name("_ns")->flatten("::");
	file << indent << "inline operator ::" << supername << "();\n";
}

void CPPFile::output_upcast(Interface *iface, Interface *super)
{
	const String &supername = *super->get_fq_name("_ns")->flatten("::");

	file << '\n' << indent << "inline "
	     << get_definition_name(iface)
	     << "::operator ::" << supername << "()\n"
	     << indent << "{\n";
	
	indent.indent_level++;
	
	// A space is added after the final '<', as GCC gives a parse
	// error otherwise.  I'm not sure exactly why...
	
	file << indent << "uintptr_t ptr = reinterpret_cast<uintptr_t>(_ptr);\n\n"
	     << indent << "if (!_ptr)\n"
	     << indent << "\treturn NULL;\n\n"
	     << indent << "ptr += _ptr->info->iface_table." 
	     << *super->get_fq_name()->flatten("_IDLNS_") 
	     << ".offset;\n"
	     << indent << "return ::" << supername << "(reinterpret_cast< ";
	
	cpp_output_name(file, super, "_i_");
	
	file << " *>(ptr));\n";
	
	indent.indent_level--;
	
	file << indent << "}\n";
}

void CPPFile::wrapper_cast_callback(Interface *super, void *arg)
{
	const callback_data *data = static_cast<const callback_data *>(arg);
	
	data->file->output_downcast(data->iface, super);
	data->file->output_upcast(data->iface, super);
}

void CPPFile::wrapper_cast_proto_callback(Interface *super, void *arg)
{
	const callback_data *data = static_cast<const callback_data *>(arg);
	
	data->file->output_downcast_proto(data->iface, super);
	data->file->output_upcast_proto(data->iface, super);
}

void CPPFile::output_wrapper(Interface *iface)
{
	const String &name = **iface->name;

	file << indent << "_i_" << name << " *_ptr;\n\n";
	
	file << indent << name << "()\n"
	     << indent << "{\n"
	     << indent << "\t_ptr = NULL;\n"
	     << indent << "}\n\n";

	file << indent << name << "(_i_" << name << " *_other)\n"
	     << indent << "{\n"
	     << indent << "\t_ptr = _other;\n"
	     << indent << "}\n\n";

	file << indent << "operator _i_" << name << " *()\n"
	     << indent << "{\n"
	     << indent << "\treturn _ptr;\n"
	     << indent << "}\n";

	do_extra_newline = true;
	
	callback_data data = { this, iface, true };

	iface->for_each_super<wrapper_cast_proto_callback>(&data);

	output_methods(iface, iface, true);
	iface->for_each_super<wrapper_method_callback>(&data);
}

void CPPFile::output_casts(Interface *iface)
{
	callback_data data = { this, iface };

	iface->for_each_super<wrapper_cast_callback>(&data);
}

void CPPFile::wrapper_method_callback(Interface *super, void *arg)
{
	const callback_data *data = static_cast<const callback_data *>(arg);

	data->file->output_methods(data->iface, super, data->prototype);
}

void CPPFile::output_method_defs(Interface *iface)
{
	output_methods(iface, iface, false);

	callback_data data = { this, iface, false };
	iface->for_each_super<wrapper_method_callback>(&data);
}
