// Bitmap Tree Allocator
//
// Written by Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// This software is in the public domain.

#ifndef _UTIL_BMAPTREE_H
#define _UTIL_BMAPTREE_H

#include <assert.h>
#include <stddef.h>
#include <stdint.h>

#include <lowlevel/types.h>
#include <lowlevel/bitops.h>

#include <orb.h>
#include <System/Exceptions.h>

namespace Util {
	// Key must be an unsigned integer type.
	template <typename Key>
	class BitmapTree {
		enum {
			node_size = sizeof(unsigned long) * 8,
			radix_bits = _LL_LONG_LOGBYTES + 3,
			key_bits = sizeof(Key) * 8,
		};
		
		struct Node {
			unsigned long index_bmap;
			
			Node() : index_bmap(0)
			{
			}
		};
		
		struct DirNode : public Node {
			Node *nodes[node_size];
			
			DirNode()
			{
				memset(nodes, 0, sizeof(nodes));
			}
		};

		Node *toplevel;
		int toplevel_shift;
		
		struct LeafNode : public Node {
			unsigned long nodes[node_size];
			
			LeafNode()
			{
				memset(nodes, 0, sizeof(nodes));
			}
		};

		static uint key_to_offset(Key key, int shift)
		{
			return (key >> shift) & (node_size - 1);
		}

	public:
		// max_key must be positive.  Modifications must be synchronized with
		// alloc() and free(), but are otherwise unrestricted (including
		// reducing it below already allocated keys).

		Key max_key;
	
	private:
		Key rec_alloc(Node *node, int shift, bool *full)
		{
			assert(~node->index_bmap != 0);
			int off = ll_ffs(!node->index_bmap);
			Key key = ((Key)off) << shift;
		
			if (shift > radix_bits) {
				assert(shift >= radix_bits * 2);
				
				if (key >= max_key)
					throw_idl(::System::Exceptions::Std::OutOfSpace,
					          ::System::RunTime::nullarray);
				
				DirNode *dn = static_cast<DirNode *>(node);
				Node *new_node = dn->nodes[off];
				
				if (unlikely(!new_node)) {
					if (shift == radix_bits * 2)
						new_node = new LeafNode;
					else
						new_node = new DirNode;
					
					dn->nodes[off] = new_node;
				}

				bool sub_full = false;
				key += rec_alloc(new_node, shift - radix_bits, &sub_full);
				
				if (sub_full)
					dn->index_bmap |= 1UL << off;
			} else {
				assert(shift == radix_bits);
	
				LeafNode *ln = static_cast<LeafNode *>(node);
				assert(~ln->nodes[off] != 0);
				
				int leaf_bit = ll_ffs(ln->nodes[off]);
				key |= ((Key)1) << leaf_bit;

				if (key >= max_key)
					throw_idl(::System::Exceptions::Std::OutOfSpace,
					          ::System::RunTime::nullarray);
				
				ln->nodes[off] |= 1UL << leaf_bit;
			
				if (~ln->nodes[off] == 0)
					ln->index_bmap |= 1UL << off;
			}

			if (~node->index_bmap == 0 && full)
				*full = true;
			
			return key;
		}
		
	public:
		// Throws System::Exceptions::Std::OutOfSpace on failure
		Key alloc()
		{
			if (unlikely(~toplevel->index_bmap == 0)) {
				int new_shift = toplevel_shift + radix_bits;
			
				if (new_shift >= key_bits ||
				    (((Key)1) << new_shift) >= max_key)
					throw_idl(::System::Exceptions::Std::OutOfSpace,
					          ::System::RunTime::nullarray);
				
				DirNode *dn = new DirNode;
				dn->nodes[0] = toplevel;
				toplevel_shift = new_shift;
				toplevel = dn;
			}
			
			return rec_alloc(toplevel, toplevel_shift, NULL);
		}
		
		// Throws System::InvalidArgument if key is already free
		void free(Key key)
		{
			Node *node = toplevel;
			int shift = toplevel_shift;
			int max_bits = toplevel_shift + radix_bits;
			
			if (max_bits < key_bits && key >= (((Key)1) << max_bits))
				throw_idl(::System::Exceptions::Std::InvalidArgument, 0,
				          ::System::RunTime::nullarray);
				
			while (shift > radix_bits) {
				assert(shift >= radix_bits * 2);

				int off = key_to_offset(key, shift);
				node->index_bmap &= ~(1UL << off);
				
				DirNode *dn = static_cast<DirNode *>(node);
				node = dn->nodes[off];
				shift -= radix_bits;
			}
			
			assert(shift == radix_bits);

			int off = key_to_offset(key, radix_bits);
			node->index_bmap &= ~(1UL << off);
			
			LeafNode *ln = static_cast<LeafNode *>(node);
			int bit = key_to_offset(key, 0);

			if (!(ln->nodes[off] & (1UL << bit)))
				throw_idl(::System::Exceptions::Std::InvalidArgument, 0,
				          ::System::RunTime::nullarray);

			ln->nodes[off] &= ~(1UL << bit);
		}
		
		BitmapTree()
		{
			max_key = ~(Key)0;
			toplevel_shift = radix_bits;
			toplevel = new LeafNode;
		}
	};
}

#endif
